/*-
 * Copyright (c) 2008 Weongyo Jeong <weongyo@freebsd.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer,
 *    without modification.
 * 2. Redistributions in binary form must reproduce at minimum a disclaimer
 *    similar to the "NO WARRANTY" disclaimer below ("Disclaimer") and any
 *    redistribution must be conditioned upon including a substantially
 *    similar Disclaimer requirement for further binary redistribution.
 *
 * NO WARRANTY
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF NONINFRINGEMENT, MERCHANTIBILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGES.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

/*
 * the Sonics Silicon Backplane(SSB) driver.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/module.h>
#include <sys/kernel.h>
#include <sys/endian.h>
#include <sys/errno.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/bus.h>
#include <sys/rman.h>
#include <sys/socket.h>

#include <net/if.h>
#include <net/if_media.h>
#include <net/if_arp.h>

#include <dev/pci/pcivar.h>
#include <dev/pci/pcireg.h>

#include <dev/ssb/ssb.h>

#ifdef SSB_DEBUG
enum {
	SSB_DEBUG_SCAN		= 0x00000001,	/* scan */
	SSB_DEBUG_PMU		= 0x00000002,	/* PMU */
	SSB_DEBUG_PLL		= 0x00000004,	/* PLL */
	SSB_DEBUG_SWITCHCORE	= 0x00000008,	/* switching core */
	SSB_DEBUG_SPROM		= 0x00000010,	/* SPROM */
	SSB_DEBUG_CORE		= 0x00000020,	/* handling cores */
	SSB_DEBUG_ANY		= 0xffffffff
};
#define DPRINTF(ssb, m, fmt, ...) do {			\
	if (ssb->ssb_debug & (m))			\
		printf(fmt, __VA_ARGS__);		\
} while (0)
#else
#define DPRINTF(ssb, m, fmt, ...) do { (void) ssb; } while (0)
#endif
#define	N(a)			(sizeof(a) / sizeof(a[0]))

static void	ssb_pci_gpio(struct ssb_softc *, uint32_t, int);
static void	ssb_scan(struct ssb_softc *);
static int	ssb_switchcore(struct ssb_softc *, uint8_t);
static int	ssb_pci_switchcore_sub(struct ssb_softc *, uint8_t);
static uint32_t	ssb_scan_read32(struct ssb_softc *, uint8_t, uint16_t);
static uint16_t	ssb_dev2chipid(struct ssb_softc *);
static uint8_t	ssb_chipid2ncores(device_t, uint16_t);
static uint16_t	ssb_pci_read_2(struct ssb_dev_softc *, uint16_t);
static uint32_t	ssb_pci_read_4(struct ssb_dev_softc *, uint16_t);
static void	ssb_pci_write_2(struct ssb_dev_softc *, uint16_t, uint16_t);
static void	ssb_pci_write_4(struct ssb_dev_softc *, uint16_t, uint32_t);
static void	ssb_cc_clock(struct ssb_cc *,
		    enum ssb_clock);
static void	ssb_cc_pmu_init(struct ssb_cc *);
static void	ssb_cc_power_init(struct ssb_cc *);
static void	ssb_cc_powerup_delay(struct ssb_cc *);
static int	ssb_cc_clockfreq(struct ssb_cc *, int);
static void	ssb_cc_pmu1_pll0_init(struct ssb_cc *, uint32_t);
static void	ssb_cc_pmu0_pll0_init(struct ssb_cc *, uint32_t);
static enum ssb_clksrc ssb_cc_clksrc(struct ssb_cc *);
static const struct ssb_cc_pmu1_plltab *ssb_cc_pmu1_plltab_find(uint32_t);
static uint32_t	ssb_cc_pll_read(struct ssb_cc *, uint32_t);
static void	ssb_cc_pll_write(struct ssb_cc *, uint32_t,
		    uint32_t);
static const struct ssb_cc_pmu0_plltab *ssb_cc_pmu0_plltab_find_entry(uint32_t);
static int	ssb_pci_sprom(struct ssb_softc *, struct ssb_sprom *);
static void	ssb_pci_get_boardinfo(struct ssb_softc *,
		    struct ssb_board *);
static int	ssb_sprom_read(struct ssb_softc *, uint16_t *, uint16_t);
static int	sprom_check_crc(const uint16_t *, size_t);
static int	ssb_sprom(struct ssb_softc *, struct ssb_sprom *,
		    const uint16_t *, uint16_t);
static uint8_t	ssb_crc8(uint8_t, uint8_t);
static void	ssb_sprom_r123(struct ssb_sprom *, const uint16_t *);
static void	ssb_sprom_r45(struct ssb_sprom *, const uint16_t *);
static void	ssb_sprom_r8(struct ssb_sprom *, const uint16_t *);
static int8_t	ssb_sprom_r123_antgain(uint8_t, const uint16_t *, uint16_t,
		    uint16_t);
static uint32_t	ssb_tmslow_reject_bitmask(struct ssb_dev_softc *);
static uint32_t	ssb_pcicore_read32(struct ssb_pci *, uint16_t);
static void	ssb_pcicore_write32(struct ssb_pci *, uint16_t, uint32_t);
static uint32_t	ssb_pcie_read(struct ssb_pci *, uint32_t);
static void	ssb_pcie_write(struct ssb_pci *, uint32_t, uint32_t);
static void	ssb_pcie_mdio_write(struct ssb_pci *, uint8_t, uint8_t,
		    uint16_t);
static void	ssb_pci_read_multi_1(struct ssb_dev_softc *, void *, size_t,
		    uint16_t);
static void	ssb_pci_read_multi_2(struct ssb_dev_softc *, void *, size_t,
		    uint16_t);
static void	ssb_pci_read_multi_4(struct ssb_dev_softc *, void *, size_t,
		    uint16_t);
static void	ssb_pci_write_multi_1(struct ssb_dev_softc *, const void *,
		    size_t, uint16_t);
static void	ssb_pci_write_multi_2(struct ssb_dev_softc *, const void *,
		    size_t, uint16_t);
static void	ssb_pci_write_multi_4(struct ssb_dev_softc *, const void *,
		    size_t, uint16_t);
#ifdef SSB_DEBUG
static const char *ssb_core_name(uint16_t);
#endif
static void	ssb_pcicore_init(struct ssb_pci *);
device_t	ssb_add_child(device_t, struct ssb_softc *, int, const char *,
		    int);

static const struct ssb_bus_ops ssb_pci_ops = {
	.read16		= ssb_pci_read_2,
	.read32		= ssb_pci_read_4,
	.write16	= ssb_pci_write_2,
	.write32	= ssb_pci_write_4,
	.read_multi_1	= ssb_pci_read_multi_1,
	.read_multi_2	= ssb_pci_read_multi_2,
	.read_multi_4	= ssb_pci_read_multi_4,
	.write_multi_1	= ssb_pci_write_multi_1,
	.write_multi_2	= ssb_pci_write_multi_2,
	.write_multi_4	= ssb_pci_write_multi_4,
};

static const struct ssb_cc_pmu_res_updown ssb_cc_pmu_4325_updown[] =
    SSB_CC_PMU_4325_RES_UPDOWN;
static const struct ssb_cc_pmu_res_depend ssb_cc_pmu_4325_depend[] =
    SSB_CC_PMU_4325_RES_DEPEND;
static const struct ssb_cc_pmu_res_updown ssb_cc_pmu_4328_updown[] =
    SSB_CC_PMU_4328_RES_UPDOWN;
static const struct ssb_cc_pmu_res_depend ssb_cc_pmu_4328_depend[] =
    SSB_CC_PMU_4328_RES_DEPEND;
static const struct ssb_cc_pmu0_plltab ssb_cc_pmu0_plltab[] =
    SSB_CC_PMU0_PLLTAB_ENTRY;
static const struct ssb_cc_pmu1_plltab ssb_cc_pmu1_plltab[] =
    SSB_CC_PMU1_PLLTAB_ENTRY;

int
ssb_attach(struct ssb_softc *ssb)
{
	struct ssb_cc *scc;
	struct ssb_info iv;
	int error;

	KASSERT(ssb->ssb_type == SSB_TYPE_PCI,
	    ("unsupported BUS type (%#x)", ssb->ssb_type));

	ssb->ssb_ops = &ssb_pci_ops;

	ssb_pci_gpio(ssb, SSB_GPIO_CRYSTAL | SSB_GPIO_PLL, 1);
	ssb_scan(ssb);

	/* XXX init PCI or PCMCIA host devices */

	ssb_powerup(ssb, 0);

	/* init ChipCommon */
	scc = &ssb->ssb_cc;
	if (scc->scc_dev != NULL) {
		ssb_cc_pmu_init(scc);
		ssb_cc_power_init(scc);
		ssb_cc_clock(scc, SSB_CLOCK_FAST);
		ssb_cc_powerup_delay(scc);
	}

	/* fetch various internal informations for PCI */
	memset(&iv, 0, sizeof(iv));
	ssb_pci_get_boardinfo(ssb, &iv.boardinfo);
	error = ssb_pci_sprom(ssb, &iv.sprom);
	if (error) {
		ssb_powerdown(ssb);
		return (error);
	}
	bcopy(&iv.boardinfo, &ssb->ssb_board, sizeof(iv.boardinfo));
	bcopy(&iv.sprom, &ssb->ssb_sprom, sizeof(iv.sprom));

	ssb_powerdown(ssb);
	return (0);
}

int
ssb_detach(struct ssb_softc *ssb)
{
	device_t *devlistp;
	int devcnt, error = 0, i;

	error = device_get_children(ssb->ssb_dev, &devlistp, &devcnt);
	if (error != 0)
		return (0);

	for ( i = 0 ; i < devcnt ; i++)
		device_delete_child(ssb->ssb_dev, devlistp[i]);
	free(devlistp, M_TEMP);
	return (0);
}

static void
ssb_pci_gpio(struct ssb_softc *ssb, uint32_t what, int on)
{
	uint32_t in, out;
	uint16_t status;

	if (ssb->ssb_type != SSB_TYPE_PCI)
		return;

	out = pci_read_config(ssb->ssb_dev, SSB_GPIO_OUT, 4);
	if (on == 0) {
		if (what & SSB_GPIO_PLL)
			out |= SSB_GPIO_PLL;
		if (what & SSB_GPIO_CRYSTAL)
			out &= ~SSB_GPIO_CRYSTAL;
		pci_write_config(ssb->ssb_dev, SSB_GPIO_OUT, out, 4);
		pci_write_config(ssb->ssb_dev, SSB_GPIO_OUT_EN,
		    pci_read_config(ssb->ssb_dev,
			SSB_GPIO_OUT_EN, 4) | what, 4);
		return;
	}

	in = pci_read_config(ssb->ssb_dev, SSB_GPIO_IN, 4);
	if ((in & SSB_GPIO_CRYSTAL) != SSB_GPIO_CRYSTAL) {
		if (what & SSB_GPIO_CRYSTAL) {
			out |= SSB_GPIO_CRYSTAL;
			if (what & SSB_GPIO_PLL)
				out |= SSB_GPIO_PLL;
			pci_write_config(ssb->ssb_dev, SSB_GPIO_OUT, out, 4);
			pci_write_config(ssb->ssb_dev,
			    SSB_GPIO_OUT_EN, pci_read_config(ssb->ssb_dev,
				SSB_GPIO_OUT_EN, 4) | what, 4);
			DELAY(1000);
		}
		if (what & SSB_GPIO_PLL) {
			out &= ~SSB_GPIO_PLL;
			pci_write_config(ssb->ssb_dev, SSB_GPIO_OUT, out, 4);
			DELAY(5000);
		}
	}

	status = pci_read_config(ssb->ssb_dev, PCIR_STATUS, 2);
	status &= ~PCIM_STATUS_STABORT;
	pci_write_config(ssb->ssb_dev, PCIR_STATUS, status, 2);
}

static void
ssb_scan(struct ssb_softc *ssb)
{
	struct ssb_dev_softc *sd;
	uint32_t idhi, tmp;
	int base, dev_i = 0, error, i, is_pcie, n_80211 = 0, n_cc = 0,
	    n_pci = 0;

	KASSERT(ssb->ssb_type == SSB_TYPE_PCI,
	    ("unsupported BUS type (%#x)", ssb->ssb_type));

	ssb->ssb_ndevs = 0;
	error = ssb_switchcore(ssb, 0); /* need the first core */
	if (error)
		return;

	idhi = ssb_scan_read32(ssb, 0, SSB_IDHIGH);
	if (SSB_IDHIGH_CORECODE(idhi) == SSB_DEV_CHIPCOMMON) {
		tmp = ssb_scan_read32(ssb, 0, SSB_CC_CHIPID);
		ssb->ssb_chipid = SSB_CC_ID(tmp);
		ssb->ssb_chiprev = SSB_CC_REV(tmp);
		ssb->ssb_chippkg = SSB_CC_PKG(tmp);
		if (SSB_IDHIGH_REV(idhi) >= 4)
			ssb->ssb_ndevs = SSB_CC_NCORES(tmp);
		ssb->ssb_cc.scc_caps = ssb_scan_read32(ssb, 0,
		    SSB_CC_CAPS);
	} else {
		if (ssb->ssb_type == SSB_TYPE_PCI) {
			ssb->ssb_chipid = ssb_dev2chipid(ssb);
			ssb->ssb_chiprev = pci_read_config(ssb->ssb_dev,
			    PCIR_REVID, 2);
			ssb->ssb_chippkg = 0;
		} else {
			ssb->ssb_chipid = 0x4710;
			ssb->ssb_chiprev = 0;
			ssb->ssb_chippkg = 0;
		}
	}
	if (ssb->ssb_ndevs == 0)
		ssb->ssb_ndevs = ssb_chipid2ncores(ssb->ssb_dev,
		    ssb->ssb_chipid);
	if (ssb->ssb_ndevs > SSB_MAX_CORES) {
		device_printf(ssb->ssb_dev,
		    "too many ssb cores (max %d %d)\n",
		    SSB_MAX_CORES, ssb->ssb_ndevs);
		return;
	}

	/* looking basic information about each cores/devices */
	for (i = 0; i < ssb->ssb_ndevs; i++) {
		error = ssb_switchcore(ssb, i);
		if (error)
			return;
		sd = &(ssb->ssb_devs[dev_i]);
		idhi = ssb_scan_read32(ssb, i, SSB_IDHIGH);
		sd->sd_bus = ssb;
		sd->sd_id.cid = SSB_IDHIGH_CORECODE(idhi);
		sd->sd_id.rev = SSB_IDHIGH_REV(idhi);
		sd->sd_id.vendor = SSB_IDHIGH_VENDOR(idhi);
		sd->sd_ops = ssb->ssb_ops;
		sd->sd_coreidx = i;

		DPRINTF(ssb, SSB_DEBUG_SCAN,
		    "core %d (%s) found (cc %#xrev %#x vendor %#x)\n",
		    i, ssb_core_name(sd->sd_id.cid),
		    sd->sd_id.cid, sd->sd_id.rev, sd->sd_id.vendor);

		switch (sd->sd_id.cid) {
		case SSB_DEV_CHIPCOMMON:
			n_cc++;
			if (n_cc > 1) {
				device_printf(ssb->ssb_dev,
				    "warn: multiple ChipCommon\n");
				break;
			}
			ssb->ssb_cc.scc_dev = sd;
			break;
		case SSB_DEV_80211:
			n_80211++;
			if (n_80211 > 1) {
				device_printf(ssb->ssb_dev,
				    "warn: multiple 802.11 core\n");
				continue;
			}
			break;
		case SSB_DEV_PCI:
		case SSB_DEV_PCIE:
			n_pci++;
			error = pci_find_extcap(ssb->ssb_dev, PCIY_EXPRESS,
			    &base);
			is_pcie = (error == 0) ? 1 : 0;

			if (n_pci > 1) {
				device_printf(ssb->ssb_dev,
				    "warn: multiple PCI(E) cores\n");
				break;
			}
			if (sd->sd_id.cid == SSB_DEV_PCI && is_pcie == 1)
				continue;
			if (sd->sd_id.cid == SSB_DEV_PCIE && is_pcie == 0)
				continue;
			ssb->ssb_pci.spc_dev = sd;
			break;
		case SSB_DEV_V90:
		case SSB_DEV_PCMCIA:
			break;
		default:
			device_printf(ssb->ssb_dev,
			    "unsupportted coreid %#x\n", sd->sd_id.cid);
			break;
		}
		dev_i++;
	}
	ssb->ssb_ndevs = dev_i;
}

static int
ssb_switchcore(struct ssb_softc *ssb, uint8_t idx)
{

	switch (ssb->ssb_type) {
	case SSB_TYPE_PCI:
		return ssb_pci_switchcore_sub(ssb, idx);
	default:
		KASSERT(0 == 1,
		    ("%s: unsupportted bustype %#x", __func__,
		    ssb->ssb_type));
	}
	return (0);
}

static int
ssb_pci_switchcore_sub(struct ssb_softc *ssb, uint8_t idx)
{
#define RETRY_MAX	50
	int i;
	uint32_t dir;

	dir = SSB_REGWIN(idx);

	for (i = 0; i < RETRY_MAX; i++) {
		pci_write_config(ssb->ssb_dev, SSB_BAR0, dir, 4);
		if (pci_read_config(ssb->ssb_dev, SSB_BAR0, 4) == dir)
			return (0);
		DELAY(10);
	}
	return (ENODEV);
#undef RETRY_MAX
}

static int
ssb_pci_switchcore(struct ssb_softc *ssb, struct ssb_dev_softc *sd)
{
	int error;

	DPRINTF(ssb, SSB_DEBUG_SWITCHCORE, "Switching to %s core, index %d\n",
	    ssb_core_name(sd->sd_id.cid), sd->sd_coreidx);

	error = ssb_pci_switchcore_sub(ssb, sd->sd_coreidx);
	if (error == 0)
		ssb->ssb_curdev = sd;

	return (error);
}

static uint32_t
ssb_scan_read32(struct ssb_softc *ssb, uint8_t coreidx,
    uint16_t offset)
{

	(void)coreidx;
	KASSERT(ssb->ssb_type == SSB_TYPE_PCI,
	    ("unsupported BUS type (%#x)", ssb->ssb_type));

	return SSB_READ_4(ssb, offset);
}

static uint16_t
ssb_dev2chipid(struct ssb_softc *ssb)
{
	uint16_t chipid = 0;

	switch (ssb->ssb_pci_did) {
	case 0x4301:
		chipid = 0x4301;
		break;
	case 0x4305:
	case 0x4306:
	case 0x4307:
		chipid = 0x4307;
		break;
	case 0x4403:
		chipid = 0x4402;
		break;
	case 0x4610:
	case 0x4611:
	case 0x4612:
	case 0x4613:
	case 0x4614:
	case 0x4615:
		chipid = 0x4610;
		break;
	case 0x4710:
	case 0x4711:
	case 0x4712:
	case 0x4713:
	case 0x4714:
	case 0x4715:
		chipid = 0x4710;
		break;
	case 0x4320:
	case 0x4321:
	case 0x4322:
	case 0x4323:
	case 0x4324:
	case 0x4325:
		chipid = 0x4309;
		break;
	case PCI_DEVICE_ID_BCM4401:
	case PCI_DEVICE_ID_BCM4401B0:
	case PCI_DEVICE_ID_BCM4401B1:
		chipid = 0x4401;
		break;
	default:
		device_printf(ssb->ssb_dev, "unknown PCI did (%d)\n",
		    ssb->ssb_pci_did);
	}

	return chipid;
}

static uint8_t
ssb_chipid2ncores(device_t dev, uint16_t chipid)
{
	switch (chipid) {
	case 0x4401:
	case 0x4402:
		return 3;
	case 0x4301:
	case 0x4307:
		return 5;
	case 0x4306:
		return 6;
	case 0x5365:
		return 7;
	case 0x4310:
		return 8;
	case 0x4710:
	case 0x4610:
	case 0x4704:
		return 9;
	default:
		device_printf(dev, "unknown the chipset ID %#x\n", chipid);
	}

	return 1;
}

#ifdef SSB_DEBUG
static const char *
ssb_core_name(uint16_t coreid)
{

	switch (coreid) {
	case SSB_DEV_CHIPCOMMON:
		return "ChipCommon";
	case SSB_DEV_ILINE20:
		return "ILine 20";
	case SSB_DEV_SDRAM:
		return "SDRAM";
	case SSB_DEV_PCI:
		return "PCI";
	case SSB_DEV_MIPS:
		return "MIPS";
	case SSB_DEV_ETHERNET:
		return "Fast Ethernet";
	case SSB_DEV_V90:
		return "V90";
	case SSB_DEV_USB11_HOSTDEV:
		return "USB 1.1 Hostdev";
	case SSB_DEV_ADSL:
		return "ADSL";
	case SSB_DEV_ILINE100:
		return "ILine 100";
	case SSB_DEV_IPSEC:
		return "IPSEC";
	case SSB_DEV_PCMCIA:
		return "PCMCIA";
	case SSB_DEV_INTERNAL_MEM:
		return "Internal Memory";
	case SSB_DEV_MEMC_SDRAM:
		return "MEMC SDRAM";
	case SSB_DEV_EXTIF:
		return "EXTIF";
	case SSB_DEV_80211:
		return "IEEE 802.11";
	case SSB_DEV_MIPS_3302:
		return "MIPS 3302";
	case SSB_DEV_USB11_HOST:
		return "USB 1.1 Host";
	case SSB_DEV_USB11_DEV:
		return "USB 1.1 Device";
	case SSB_DEV_USB20_HOST:
		return "USB 2.0 Host";
	case SSB_DEV_USB20_DEV:
		return "USB 2.0 Device";
	case SSB_DEV_SDIO_HOST:
		return "SDIO Host";
	case SSB_DEV_ROBOSWITCH:
		return "Roboswitch";
	case SSB_DEV_PARA_ATA:
		return "PATA";
	case SSB_DEV_SATA_XORDMA:
		return "SATA XOR-DMA";
	case SSB_DEV_ETHERNET_GBIT:
		return "GBit Ethernet";
	case SSB_DEV_PCIE:
		return "PCI-E";
	case SSB_DEV_MIMO_PHY:
		return "MIMO PHY";
	case SSB_DEV_SRAM_CTRLR:
		return "SRAM Controller";
	case SSB_DEV_MINI_MACPHY:
		return "Mini MACPHY";
	case SSB_DEV_ARM_1176:
		return "ARM 1176";
	case SSB_DEV_ARM_7TDMI:
		return "ARM 7TDMI";
	}
	return "unknown";
}
#endif

static uint16_t
ssb_pci_read_2(struct ssb_dev_softc *sd, uint16_t offset)
{
	struct ssb_softc *ssb = sd->sd_bus;
	uint16_t value;

	if (ssb->ssb_curdev != sd && ssb_pci_switchcore(ssb, sd) != 0)
		return (0xffff);

	value = SSB_READ_2(ssb, offset);
	return (value);
}

static uint32_t
ssb_pci_read_4(struct ssb_dev_softc *sd, uint16_t offset)
{
	struct ssb_softc *ssb = sd->sd_bus;
	uint32_t value;

	if (ssb->ssb_curdev != sd && ssb_pci_switchcore(ssb, sd) != 0)
		return (0xffff);

	value = SSB_READ_4(ssb, offset);
	return (value);
}

static void
ssb_pci_write_2(struct ssb_dev_softc *sd, uint16_t offset, uint16_t value)
{
	struct ssb_softc *ssb = sd->sd_bus;

	if (ssb->ssb_curdev != sd && ssb_pci_switchcore(ssb, sd) != 0)
		return;

	SSB_WRITE_2(ssb, offset, value);
}

static void
ssb_pci_write_4(struct ssb_dev_softc *sd, uint16_t offset, uint32_t value)
{
	struct ssb_softc *ssb = sd->sd_bus;

	if (ssb->ssb_curdev != sd && ssb_pci_switchcore(ssb, sd) != 0)
		return;

	SSB_WRITE_4(ssb, offset, value);
}

static void
ssb_pci_read_multi_1(struct ssb_dev_softc *sd, void *buffer, size_t count,
    uint16_t offset)
{
	struct ssb_softc *ssb = sd->sd_bus;

	if (ssb->ssb_curdev != sd && ssb_pci_switchcore(ssb, sd) != 0) {
		memset(buffer, 0xff, count);
		return;
	}

	SSB_READ_MULTI_1(ssb, offset, buffer, count);
}

static void
ssb_pci_read_multi_2(struct ssb_dev_softc *sd, void *buffer, size_t count,
    uint16_t offset)
{
	struct ssb_softc *ssb = sd->sd_bus;

	if (ssb->ssb_curdev != sd && ssb_pci_switchcore(ssb, sd) != 0) {
		memset(buffer, 0xff, count);
		return;
	}

	KASSERT(!(count & 1), ("%s:%d: fail", __func__, __LINE__));
	SSB_READ_MULTI_2(ssb, offset, buffer, count >> 1);
}

static void
ssb_pci_read_multi_4(struct ssb_dev_softc *sd, void *buffer, size_t count,
    uint16_t offset)
{
	struct ssb_softc *ssb = sd->sd_bus;

	if (ssb->ssb_curdev != sd && ssb_pci_switchcore(ssb, sd) != 0) {
		memset(buffer, 0xff, count);
		return;
	}

	KASSERT(!(count & 3), ("%s:%d: fail", __func__, __LINE__));
	SSB_READ_MULTI_4(ssb, offset, buffer, count >> 2);
}

static void
ssb_pci_write_multi_1(struct ssb_dev_softc *sd, const void *buffer,
    size_t count, uint16_t offset)
{
	struct ssb_softc *ssb = sd->sd_bus;

	if (ssb->ssb_curdev != sd && ssb_pci_switchcore(ssb, sd) != 0)
		return;

	SSB_WRITE_MULTI_1(ssb, offset, buffer, count);
}

static void
ssb_pci_write_multi_2(struct ssb_dev_softc *sd, const void *buffer,
    size_t count, uint16_t offset)
{
	struct ssb_softc *ssb = sd->sd_bus;

	if (ssb->ssb_curdev != sd && ssb_pci_switchcore(ssb, sd) != 0)
		return;

	KASSERT(!(count & 1), ("%s:%d: fail", __func__, __LINE__));
	SSB_WRITE_MULTI_2(ssb, offset, buffer, count >> 1);
}

static void
ssb_pci_write_multi_4(struct ssb_dev_softc *sd, const void *buffer,
    size_t count, uint16_t offset)
{
	struct ssb_softc *ssb = sd->sd_bus;

	if (ssb->ssb_curdev != sd && ssb_pci_switchcore(ssb, sd) != 0)
		return;

	KASSERT(!(count & 3), ("%s:%d: fail", __func__, __LINE__));
	SSB_WRITE_MULTI_4(ssb, offset, buffer, count >> 2);
}

void
ssb_powerup(struct ssb_softc *ssb, int dynamic)
{

	ssb_pci_gpio(ssb, SSB_GPIO_CRYSTAL | SSB_GPIO_PLL, 1);
	ssb_cc_clock(&ssb->ssb_cc,
	    (dynamic != 0) ? SSB_CLOCK_DYNAMIC : SSB_CLOCK_FAST);
}

static void
ssb_cc_clock(struct ssb_cc *scc, enum ssb_clock clock)
{
	struct ssb_dev_softc *sd = scc->scc_dev;
	struct ssb_softc *ssb;
	uint32_t tmp;

	if (sd == NULL)
		return;
	ssb = sd->sd_bus;
	/*
	 * chipcommon < r6 (no dynamic clock control)
	 * chipcommon >= r10 (unknown)
	 */
	if (sd->sd_id.rev < 6 || sd->sd_id.rev >= 10 ||
	    (scc->scc_caps & SSB_CC_CAPS_PWCTL) == 0)
		return;

	switch (clock) {
	case SSB_CLOCK_DYNAMIC:
		tmp = SSB_CC_READ32(scc, SSB_CC_CLKSLOW) &
		    ~(SSB_CC_CLKSLOW_ENXTAL | SSB_CC_CLKSLOW_FSLOW |
		    SSB_CC_CLKSLOW_IPLL);
		if ((tmp & SSB_CC_CLKSLOW_SRC) != SSB_CC_CLKSLOW_SRC_CRYSTAL)
			tmp |= SSB_CC_CLKSLOW_ENXTAL;
		SSB_CC_WRITE32(scc, SSB_CC_CLKSLOW, tmp);
		if (tmp & SSB_CC_CLKSLOW_ENXTAL)
			ssb_pci_gpio(ssb, SSB_GPIO_CRYSTAL, 0);
		break;
	case SSB_CLOCK_SLOW:
		SSB_CC_WRITE32(scc, SSB_CC_CLKSLOW,
		    SSB_CC_READ32(scc, SSB_CC_CLKSLOW) |
		    SSB_CC_CLKSLOW_FSLOW);
		break;
	case SSB_CLOCK_FAST:
		/* crystal on */
		ssb_pci_gpio(ssb, SSB_GPIO_CRYSTAL, 1);
		SSB_CC_WRITE32(scc, SSB_CC_CLKSLOW,
		    (SSB_CC_READ32(scc, SSB_CC_CLKSLOW) |
			SSB_CC_CLKSLOW_IPLL) & ~SSB_CC_CLKSLOW_FSLOW);
		break;
	default:
		KASSERT(0 == 1,
		    ("%s: unsupported clock %#x", __func__, clock));
	}
}

uint16_t
ssb_read_2(struct ssb_dev_softc *sd, uint16_t offset)
{

	return sd->sd_ops->read16(sd, offset);
}

uint32_t
ssb_read_4(struct ssb_dev_softc *sd, uint16_t offset)
{

	return sd->sd_ops->read32(sd, offset);
}

void
ssb_write_2(struct ssb_dev_softc *sd, uint16_t offset, uint16_t value)
{

	sd->sd_ops->write16(sd, offset, value);
}

void
ssb_write_4(struct ssb_dev_softc *sd, uint16_t offset, uint32_t value)
{

	sd->sd_ops->write32(sd, offset, value);
}

void
ssb_read_multi_1(struct ssb_dev_softc *sd, void *buffer, size_t count,
    uint16_t offset)
{

	sd->sd_ops->read_multi_1(sd, buffer, count, offset);
}

void
ssb_read_multi_2(struct ssb_dev_softc *sd, void *buffer, size_t count,
    uint16_t offset)
{

	sd->sd_ops->read_multi_2(sd, buffer, count, offset);
}

void
ssb_read_multi_4(struct ssb_dev_softc *sd, void *buffer, size_t count,
    uint16_t offset)
{

	sd->sd_ops->read_multi_4(sd, buffer, count, offset);
}

void
ssb_write_multi_1(struct ssb_dev_softc *sd, const void *buffer,
    size_t count, uint16_t offset)
{

	sd->sd_ops->write_multi_1(sd, buffer, count, offset);
}

void
ssb_write_multi_2(struct ssb_dev_softc *sd, const void *buffer,
    size_t count, uint16_t offset)
{

	sd->sd_ops->write_multi_2(sd, buffer, count, offset);
}

void
ssb_write_multi_4(struct ssb_dev_softc *sd, const void *buffer,
    size_t count, uint16_t offset)
{

	sd->sd_ops->write_multi_4(sd, buffer, count, offset);
}

static void
ssb_cc_pmu_init(struct ssb_cc *scc)
{
	const struct ssb_cc_pmu_res_updown *updown = NULL;
	const struct ssb_cc_pmu_res_depend *depend = NULL;
	struct ssb_dev_softc *sd = scc->scc_dev;
	struct ssb_softc *ssb = sd->sd_bus;
	uint32_t min = 0, max = 0, pmucap;
	unsigned int i, updown_size, depend_size;

	if ((scc->scc_caps & SSB_CC_CAPS_PMU) == 0)
		return;

	pmucap = SSB_CC_READ32(scc, SSB_CC_PMUCAPS);
	scc->scc_pmu.rev = (pmucap & SSB_CC_PMUCAPS_REV);

	DPRINTF(ssb, SSB_DEBUG_PMU, "PMU(r%u) found (caps %#x)\n",
	    scc->scc_pmu.rev, pmucap);

	if (scc->scc_pmu.rev >= 1) {
		if (ssb->ssb_chiprev < 2 && ssb->ssb_chipid == 0x4325)
			SSB_CC_MASK32(scc, SSB_CC_PMUCTL,
			    ~SSB_CC_PMUCTL_NOILP);
		else
			SSB_CC_SET32(scc, SSB_CC_PMUCTL,
			    SSB_CC_PMUCTL_NOILP);
	}

	/* initialize PLL & PMU resources */
	switch (ssb->ssb_chipid) {
	case 0x4312:
		ssb_cc_pmu1_pll0_init(scc, 0 /* use default */);
		/* use the default: min = 0xcbb max = 0x7ffff */
		break;
	case 0x4325:
		ssb_cc_pmu1_pll0_init(scc, 0 /* use default */);

		updown = ssb_cc_pmu_4325_updown;
		updown_size = N(ssb_cc_pmu_4325_updown);
		depend = ssb_cc_pmu_4325_depend;
		depend_size = N(ssb_cc_pmu_4325_depend);

		min = (1 << SSB_CC_PMU_4325_BURST) |
		    (1 << SSB_CC_PMU_4325_LN);
		if (SSB_CC_READ32(scc, SSB_CC_CHIPSTAT) &
		    SSB_CC_CHST_4325_PMUTOP_2B)
			min |= (1 << SSB_CC_PMU_4325_CLBURST);
		max = 0xfffff;
		break;
	case 0x4328:
		ssb_cc_pmu0_pll0_init(scc, 0 /* use default */);

		updown = ssb_cc_pmu_4328_updown;
		updown_size = N(ssb_cc_pmu_4328_updown);
		depend = ssb_cc_pmu_4328_depend;
		depend_size = N(ssb_cc_pmu_4328_depend);

		min = (1 << SSB_CC_PMU_4328_EXT_SWITCH_PWM) |
			  (1 << SSB_CC_PMU_4328_BB_SWITCH_PWM) |
			  (1 << SSB_CC_PMU_4328_CRYSTAL_EN);

		max = 0xfffff;
		break;
	case 0x5354:
		ssb_cc_pmu0_pll0_init(scc, 0 /* use default */);

		max = 0xfffff;
		break;
	default:
		device_printf(ssb->ssb_dev,
		    "unknown chipid %#x for PLL & PMU init\n", ssb->ssb_chipid);
	}

	if (updown) {
		for (i = 0; i < updown_size; i++) {
			SSB_CC_WRITE32(scc, SSB_CC_PMU_TABSEL,
			    updown[i].res);
			SSB_CC_WRITE32(scc, SSB_CC_PMU_UPDNTM,
			    updown[i].updown);
		}
	}
	if (depend) {
		for (i = 0; i < depend_size; i++) {
			SSB_CC_WRITE32(scc, SSB_CC_PMU_TABSEL,
			    depend[i].res);
			switch (depend[i].task) {
			case SSB_CC_PMU_DEP_SET:
				SSB_CC_WRITE32(scc, SSB_CC_PMU_DEPMSK,
				    depend[i].depend);
				break;
			case SSB_CC_PMU_DEP_ADD:
				SSB_CC_SET32(scc, SSB_CC_PMU_DEPMSK,
				    depend[i].depend);
				break;
			case SSB_CC_PMU_DEP_REMOVE:
				SSB_CC_MASK32(scc, SSB_CC_PMU_DEPMSK,
				    ~(depend[i].depend));
				break;
			default:
				KASSERT(0 == 1,
				    ("%s:%d: assertion failed",
					__func__, __LINE__));
			}
		}
	}

	if (min)
		SSB_CC_WRITE32(scc, SSB_CC_PMU_MINRES, min);
	if (max)
		SSB_CC_WRITE32(scc, SSB_CC_PMU_MAXRES, max);
}

static void
ssb_cc_power_init(struct ssb_cc *scc)
{
	struct ssb_softc *ssb = scc->scc_dev->sd_bus;
	int maxfreq;

	if (ssb->ssb_chipid == 0x4321) {
		if (ssb->ssb_chiprev == 0)
			SSB_CC_WRITE32(scc, SSB_CC_CHIPCTL, 0x3a4);
		else if (ssb->ssb_chiprev == 1)
			SSB_CC_WRITE32(scc, SSB_CC_CHIPCTL, 0xa4);
	}

	if ((scc->scc_caps & SSB_CC_CAPS_PWCTL) == 0)
		return;

	if (scc->scc_dev->sd_id.rev >= 10)
		SSB_CC_WRITE32(scc, SSB_CC_CLKSYSCTL,
		    (SSB_CC_READ32(scc, SSB_CC_CLKSYSCTL) &
		    0xffff) | 0x40000);
	else {
		maxfreq = ssb_cc_clockfreq(scc, 1);
		SSB_CC_WRITE32(scc, SSB_CC_PLLONDELAY,
		    (maxfreq * 150 + 999999) / 1000000);
		SSB_CC_WRITE32(scc, SSB_CC_FREFSELDELAY,
		    (maxfreq * 15 + 999999) / 1000000);
	}
}

static void
ssb_cc_powerup_delay(struct ssb_cc *scc)
{
	struct ssb_softc *ssb = scc->scc_dev->sd_bus;
	int min;

	if (ssb->ssb_type != SSB_TYPE_PCI ||
	    !(scc->scc_caps & SSB_CC_CAPS_PWCTL))
		return;

	min = ssb_cc_clockfreq(scc, 0);
	scc->scc_powerup_delay =
	    (((SSB_CC_READ32(scc, SSB_CC_PLLONDELAY) + 2) * 1000000) +
	    (min - 1)) / min;
}

static int
ssb_cc_clockfreq(struct ssb_cc *scc, int max)
{
	enum ssb_clksrc src;
	int div = 1, limit = 0;

	src = ssb_cc_clksrc(scc);
	if (scc->scc_dev->sd_id.rev < 6) {
		div = (src == SSB_CC_CLKSRC_PCI) ? 64 :
		    (src == SSB_CC_CLKSRC_CRYSTAL) ? 32 : 1;	
		KASSERT(div != 1,
		    ("%s: unknown clock %d", __func__, src));
	} else if (scc->scc_dev->sd_id.rev < 10) {
		switch (src) {
		case SSB_CC_CLKSRC_CRYSTAL:
		case SSB_CC_CLKSRC_PCI:
			div = ((SSB_CC_READ32(scc, SSB_CC_CLKSLOW) >> 16) +
			    1) * 4;
			break;
		case SSB_CC_CLKSRC_LOWPW:
			break;
		}
	} else
		div = ((SSB_CC_READ32(scc, SSB_CC_CLKSYSCTL) >> 16) + 1) * 4;

	switch (src) {
	case SSB_CC_CLKSRC_CRYSTAL:
		limit = (max) ? 20200000 : 19800000;
		break;
	case SSB_CC_CLKSRC_LOWPW:
		limit = (max) ? 43000 : 25000;
		break;
	case SSB_CC_CLKSRC_PCI:
		limit = (max) ? 34000000 : 25000000;
		break;
	}

	return (limit / div);
}

static void
ssb_cc_pmu1_pll0_init(struct ssb_cc *scc, uint32_t freq)
{
	struct ssb_dev_softc *sd = scc->scc_dev;
	struct ssb_softc *ssb = sd->sd_bus;
	const struct ssb_cc_pmu1_plltab *e = NULL;
	uint32_t bufsth = 0, pll, pmu;
	unsigned int i;

	KASSERT(freq == 0, ("%s:%d: assertion vail", __func__, __LINE__));
	if (ssb->ssb_chipid == 0x4312) {
		scc->scc_pmu.freq = 20000;
		return;
	}

	e = ssb_cc_pmu1_plltab_find(SSB_CC_PMU1_DEFAULT_FREQ);
	KASSERT(e != NULL, ("%s:%d: assertion vail", __func__, __LINE__));
	scc->scc_pmu.freq = e->freq;

	pmu = SSB_CC_READ32(scc, SSB_CC_PMUCTL);
	if (SSB_CC_PMUCTL_XF_VAL(pmu) == e->xf)
		return;

	DPRINTF(ssb, SSB_DEBUG_PLL, "change PLL value to %u.%03u MHz\n",
	    (e->freq / 1000), (e->freq % 1000));

	/* turn PLL off */
	switch (ssb->ssb_chipid) {
	case 0x4325:
		bufsth = 0x222222;
		SSB_CC_MASK32(scc, SSB_CC_PMU_MINRES,
		    ~((1 << SSB_CC_PMU_4325_BBPLL_PWR) |
		      (1 << SSB_CC_PMU_4325_HT)));
		SSB_CC_MASK32(scc, SSB_CC_PMU_MAXRES,
		    ~((1 << SSB_CC_PMU_4325_BBPLL_PWR) |
		      (1 << SSB_CC_PMU_4325_HT)));
		break;
	default:
		KASSERT(0 == 1,
		    ("%s:%d: assertion failed", __func__, __LINE__));
	}
	for (i = 0; i < 1500; i++) {
		if (!(SSB_CC_READ32(scc, SSB_CC_CLKCTLSTATUS) &
		      SSB_CC_CLKCTLSTATUS_HT))
			break;
		DELAY(10);
	}
	if (SSB_CC_READ32(scc, SSB_CC_CLKCTLSTATUS) & SSB_CC_CLKCTLSTATUS_HT)
		device_printf(ssb->ssb_dev, "failed to turn PLL off!\n");

	pll = ssb_cc_pll_read(scc, SSB_CC_PMU1_PLL0);
	pll &= ~(SSB_CC_PMU1_PLL0_P1DIV | SSB_CC_PMU1_PLL0_P2DIV);
	pll |= ((uint32_t)e->p1div << 20) & SSB_CC_PMU1_PLL0_P1DIV;
	pll |= ((uint32_t)e->p2div << 24) & SSB_CC_PMU1_PLL0_P2DIV;
	ssb_cc_pll_write(scc, SSB_CC_PMU1_PLL0, pll);

	pll = ssb_cc_pll_read(scc, SSB_CC_PMU1_PLL2);
	pll &= ~(SSB_CC_PMU1_PLL2_NDIVINT | SSB_CC_PMU1_PLL2_NDIVMODE);
	pll |= ((uint32_t)e->ndiv_int << 20) & SSB_CC_PMU1_PLL2_NDIVINT;
	pll |= (1 << 17) & SSB_CC_PMU1_PLL2_NDIVMODE;
	ssb_cc_pll_write(scc, SSB_CC_PMU1_PLL2, pll);

	pll = ssb_cc_pll_read(scc, SSB_CC_PMU1_PLL3);
	pll &= ~SSB_CC_PMU1_PLL3_NDIVFRAC;
	pll |= ((uint32_t)e->ndiv_frac << 0) & SSB_CC_PMU1_PLL3_NDIVFRAC;
	ssb_cc_pll_write(scc, SSB_CC_PMU1_PLL3, pll);

	if (bufsth) {
		pll = ssb_cc_pll_read(scc, SSB_CC_PMU1_PLL5);
		pll &= ~SSB_CC_PMU1_PLL5_CLKDRV;
		pll |= (bufsth << 8) & SSB_CC_PMU1_PLL5_CLKDRV;
		ssb_cc_pll_write(scc, SSB_CC_PMU1_PLL5, pll);
	}

	pmu = SSB_CC_READ32(scc, SSB_CC_PMUCTL);
	pmu &= ~(SSB_CC_PMUCTL_ILP | SSB_CC_PMUCTL_XF);
	pmu |= ((((uint32_t)e->freq + 127) / 128 - 1) << 16) &
	    SSB_CC_PMUCTL_ILP;
	pmu |= ((uint32_t)e->xf << 2) & SSB_CC_PMUCTL_XF;
	SSB_CC_WRITE32(scc, SSB_CC_PMUCTL, pmu);
}

static void
ssb_cc_pmu0_pll0_init(struct ssb_cc *scc, uint32_t crystalfreq)
{
	struct ssb_dev_softc *sd = scc->scc_dev;
	struct ssb_softc *ssb = sd->sd_bus;
	const struct ssb_cc_pmu0_plltab *e = NULL;
	uint32_t pmu, tmp, pll;
	unsigned int i;

	if ((ssb->ssb_chipid == 0x5354) && !crystalfreq)
		crystalfreq = 25000;
	if (crystalfreq)
		e = ssb_cc_pmu0_plltab_find_entry(crystalfreq);
	if (!e)
		e = ssb_cc_pmu0_plltab_find_entry(SSB_CC_PMU0_DEFAULT_XTALFREQ);
	KASSERT(e != NULL, ("%s:%d: fail", __func__, __LINE__));
	crystalfreq = e->freq;
	scc->scc_pmu.freq = e->freq;

	pmu = SSB_CC_READ32(scc, SSB_CC_PMUCTL);
	if (((pmu & SSB_CC_PMUCTL_XF) >> 2) == e->xf)
		return;

	DPRINTF(ssb, SSB_DEBUG_PLL, "change PLL value to %u.%03u mhz\n",
	    (crystalfreq / 1000), (crystalfreq % 1000));

	KASSERT(ssb->ssb_chipid == 0x4328 || ssb->ssb_chipid == 0x5354,
	    ("%s:%d: fail", __func__, __LINE__));

	switch (ssb->ssb_chipid) {
	case 0x4328:
		SSB_CC_MASK32(scc, SSB_CC_PMU_MINRES,
		    ~(1 << SSB_CC_PMU_4328_BB_PLL_PU));
		SSB_CC_MASK32(scc, SSB_CC_PMU_MAXRES,
		    ~(1 << SSB_CC_PMU_4328_BB_PLL_PU));
		break;
	case 0x5354:
		SSB_CC_MASK32(scc, SSB_CC_PMU_MINRES,
		    ~(1 << SSB_CC_PMU_5354_BB_PLL_PU));
		SSB_CC_MASK32(scc, SSB_CC_PMU_MAXRES,
		    ~(1 << SSB_CC_PMU_5354_BB_PLL_PU));
		break;
	}
	for (i = 1500; i; i--) {
		tmp = SSB_CC_READ32(scc, SSB_CC_CLKCTLSTATUS);
		if (!(tmp & SSB_CC_CLKCTLSTATUS_HT))
			break;
		DELAY(10);
	}
	tmp = SSB_CC_READ32(scc, SSB_CC_CLKCTLSTATUS);
	if (tmp & SSB_CC_CLKCTLSTATUS_HT)
		device_printf(ssb->ssb_dev, "failed to turn PLL off!\n");

	/* set PDIV */
	pll = ssb_cc_pll_read(scc, SSB_CC_PMU0_PLL0);
	if (crystalfreq >= SSB_CC_PMU0_PLL0_PDIV_FREQ)
		pll |= SSB_CC_PMU0_PLL0_PDIV_MSK;
	else
		pll &= ~SSB_CC_PMU0_PLL0_PDIV_MSK;
	ssb_cc_pll_write(scc, SSB_CC_PMU0_PLL0, pll);

	/* set WILD */
	pll = ssb_cc_pll_read(scc, SSB_CC_PMU0_PLL1);
	pll &= ~(SSB_CC_PMU0_PLL1_STOPMOD | SSB_CC_PMU0_PLL1_IMSK |
	    SSB_CC_PMU0_PLL1_FMSK);
	pll |= ((uint32_t)e->wb_int << 28) & SSB_CC_PMU0_PLL1_IMSK;
	pll |= ((uint32_t)e->wb_frac << 8) & SSB_CC_PMU0_PLL1_FMSK;
	if (e->wb_frac == 0)
		pll |= SSB_CC_PMU0_PLL1_STOPMOD;
	ssb_cc_pll_write(scc, SSB_CC_PMU0_PLL1, pll);

	/* set WILD */
	pll = ssb_cc_pll_read(scc, SSB_CC_PMU0_PLL2);
	pll &= ~SSB_CC_PMU0_PLL2_IMSKHI;
	pll |= (((uint32_t)e->wb_int >> 4) << 0) & SSB_CC_PMU0_PLL2_IMSKHI;
	ssb_cc_pll_write(scc, SSB_CC_PMU0_PLL2, pll);

	/* set freq and divisor. */
	pmu = SSB_CC_READ32(scc, SSB_CC_PMUCTL);
	pmu &= ~SSB_CC_PMUCTL_ILP;
	pmu |= (((crystalfreq + 127) / 128 - 1) << 16) & SSB_CC_PMUCTL_ILP;
	pmu &= ~SSB_CC_PMUCTL_XF;
	pmu |= ((uint32_t)e->xf << 2) & SSB_CC_PMUCTL_XF;
	SSB_CC_WRITE32(scc, SSB_CC_PMUCTL, pmu);
}

static enum ssb_clksrc
ssb_cc_clksrc(struct ssb_cc *scc)
{
	struct ssb_dev_softc *sd = scc->scc_dev;
	struct ssb_softc *ssb = sd->sd_bus;

	if (sd->sd_id.rev < 6) {
		if (ssb->ssb_type == SSB_TYPE_PCI) {
			if (pci_read_config(ssb->ssb_dev, SSB_GPIO_OUT, 4) &
			    0x10)
				return (SSB_CC_CLKSRC_PCI);
			return (SSB_CC_CLKSRC_CRYSTAL);
		}
		if (ssb->ssb_type == SSB_TYPE_SSB ||
		    ssb->ssb_type == SSB_TYPE_PCMCIA)
			return (SSB_CC_CLKSRC_CRYSTAL);
	}
	if (sd->sd_id.rev < 10) {
		switch (SSB_CC_READ32(scc, SSB_CC_CLKSLOW) & 0x7) {
		case 0:
			return (SSB_CC_CLKSRC_LOWPW);
		case 1:
			return (SSB_CC_CLKSRC_CRYSTAL);
		case 2:
			return (SSB_CC_CLKSRC_PCI);
		default:
			break;
		}
	}

	return (SSB_CC_CLKSRC_CRYSTAL);
}

static const struct ssb_cc_pmu1_plltab *
ssb_cc_pmu1_plltab_find(uint32_t crystalfreq)
{
	const struct ssb_cc_pmu1_plltab *e;
	unsigned int i;

	for (i = 0; i < N(ssb_cc_pmu1_plltab); i++) {
		e = &ssb_cc_pmu1_plltab[i];
		if (crystalfreq == e->freq)
			return (e);
	}

	return (NULL);
}

static uint32_t
ssb_cc_pll_read(struct ssb_cc *scc, uint32_t offset)
{

	SSB_CC_WRITE32(scc, SSB_CC_PLLCTL_ADDR, offset);
	return SSB_CC_READ32(scc, SSB_CC_PLLCTL_DATA);
}

static void
ssb_cc_pll_write(struct ssb_cc *scc, uint32_t offset, uint32_t value)
{

	SSB_CC_WRITE32(scc, SSB_CC_PLLCTL_ADDR, offset);
	SSB_CC_WRITE32(scc, SSB_CC_PLLCTL_DATA, value);
}

static const struct ssb_cc_pmu0_plltab *
ssb_cc_pmu0_plltab_find_entry(uint32_t crystalfreq)
{
	const struct ssb_cc_pmu0_plltab *e;
	unsigned int i;

	for (i = 0; i < N(ssb_cc_pmu0_plltab); i++) {
		e = &ssb_cc_pmu0_plltab[i];
		if (e->freq == crystalfreq)
			return e;
	}

	return NULL;
}

static int
ssb_pci_sprom(struct ssb_softc *ssb, struct ssb_sprom *sprom)
{
	int error = ENOMEM;
	uint16_t *buf;

	buf = malloc(SSB_SPROMSIZE_R123 * sizeof(uint16_t),
	    M_DEVBUF, M_NOWAIT | M_ZERO);
	if (buf == NULL)
		return (ENOMEM);
	ssb_sprom_read(ssb, buf, SSB_SPROMSIZE_R123);
	error = sprom_check_crc(buf, ssb->ssb_spromsize);
	if (error) {
		free(buf, M_DEVBUF);
		buf = malloc(SSB_SPROMSIZE_R4 * sizeof(uint16_t),
		    M_DEVBUF, M_NOWAIT | M_ZERO);
		if (buf == NULL)
			return (ENOMEM);
		ssb_sprom_read(ssb, buf, SSB_SPROMSIZE_R4);
		error = sprom_check_crc(buf, ssb->ssb_spromsize);
		if (error)
			device_printf(ssb->ssb_dev, "warn: bad SPROM CRC\n");
	}
	error = ssb_sprom(ssb, sprom, buf, ssb->ssb_spromsize);
	free(buf, M_DEVBUF);
	return error;
}

static void
ssb_pci_get_boardinfo(struct ssb_softc *ssb, struct ssb_board *bi)
{

	bi->vendor = pci_read_config(ssb->ssb_dev, PCIR_SUBVEND_0, 2);
	bi->type = pci_read_config(ssb->ssb_dev, PCIR_SUBDEV_0, 2);
	bi->rev = pci_read_config(ssb->ssb_dev, PCIR_REVID, 2);
}

static int
ssb_sprom_read(struct ssb_softc *ssb, uint16_t *sprom, uint16_t len)
{
	int i;

	for (i = 0; i < len; i++)
		sprom[i] = SSB_READ_2(ssb, SSB_SPROM_BASE + (i * 2));

	ssb->ssb_spromsize = len;
	return (0);
}

static int
sprom_check_crc(const uint16_t *sprom, size_t size)
{
	int word;
	uint8_t crc0, crc1 = 0xff;

	crc0 = (sprom[size - 1] & SSB_SPROM_REV_CRC) >> 8;
	for (word = 0; word < size - 1; word++) {
		crc1 = ssb_crc8(crc1, sprom[word] & 0x00ff);
		crc1 = ssb_crc8(crc1, (sprom[word] & 0xff00) >> 8);
	}
	crc1 = ssb_crc8(crc1, sprom[size - 1] & 0x00ff);
	crc1 ^= 0xff;

	return ((crc0 != crc1) ? EPROTO : 0);
}

static uint8_t
ssb_crc8(uint8_t crc, uint8_t data)
{
	static const uint8_t ct[] = {
		0x00, 0xf7, 0xb9, 0x4e, 0x25, 0xd2, 0x9c, 0x6b,
		0x4a, 0xbd, 0xf3, 0x04, 0x6f, 0x98, 0xd6, 0x21,
		0x94, 0x63, 0x2d, 0xda, 0xb1, 0x46, 0x08, 0xff,
		0xde, 0x29, 0x67, 0x90, 0xfb, 0x0c, 0x42, 0xb5,
		0x7f, 0x88, 0xc6, 0x31, 0x5a, 0xad, 0xe3, 0x14,
		0x35, 0xc2, 0x8c, 0x7b, 0x10, 0xe7, 0xa9, 0x5e,
		0xeb, 0x1c, 0x52, 0xa5, 0xce, 0x39, 0x77, 0x80,
		0xa1, 0x56, 0x18, 0xef, 0x84, 0x73, 0x3d, 0xca,
		0xfe, 0x09, 0x47, 0xb0, 0xdb, 0x2c, 0x62, 0x95,
		0xb4, 0x43, 0x0d, 0xfa, 0x91, 0x66, 0x28, 0xdf,
		0x6a, 0x9d, 0xd3, 0x24, 0x4f, 0xb8, 0xf6, 0x01,
		0x20, 0xd7, 0x99, 0x6e, 0x05, 0xf2, 0xbc, 0x4b,
		0x81, 0x76, 0x38, 0xcf, 0xa4, 0x53, 0x1d, 0xea,
		0xcb, 0x3c, 0x72, 0x85, 0xee, 0x19, 0x57, 0xa0,
		0x15, 0xe2, 0xac, 0x5b, 0x30, 0xc7, 0x89, 0x7e,
		0x5f, 0xa8, 0xe6, 0x11, 0x7a, 0x8d, 0xc3, 0x34,
		0xab, 0x5c, 0x12, 0xe5, 0x8e, 0x79, 0x37, 0xc0,
		0xe1, 0x16, 0x58, 0xaf, 0xc4, 0x33, 0x7d, 0x8a,
		0x3f, 0xc8, 0x86, 0x71, 0x1a, 0xed, 0xa3, 0x54,
		0x75, 0x82, 0xcc, 0x3b, 0x50, 0xa7, 0xe9, 0x1e,
		0xd4, 0x23, 0x6d, 0x9a, 0xf1, 0x06, 0x48, 0xbf,
		0x9e, 0x69, 0x27, 0xd0, 0xbb, 0x4c, 0x02, 0xf5,
		0x40, 0xb7, 0xf9, 0x0e, 0x65, 0x92, 0xdc, 0x2b,
		0x0a, 0xfd, 0xb3, 0x44, 0x2f, 0xd8, 0x96, 0x61,
		0x55, 0xa2, 0xec, 0x1b, 0x70, 0x87, 0xc9, 0x3e,
		0x1f, 0xe8, 0xa6, 0x51, 0x3a, 0xcd, 0x83, 0x74,
		0xc1, 0x36, 0x78, 0x8f, 0xe4, 0x13, 0x5d, 0xaa,
		0x8b, 0x7c, 0x32, 0xc5, 0xae, 0x59, 0x17, 0xe0,
		0x2a, 0xdd, 0x93, 0x64, 0x0f, 0xf8, 0xb6, 0x41,
		0x60, 0x97, 0xd9, 0x2e, 0x45, 0xb2, 0xfc, 0x0b,
		0xbe, 0x49, 0x07, 0xf0, 0x9b, 0x6c, 0x22, 0xd5,
		0xf4, 0x03, 0x4d, 0xba, 0xd1, 0x26, 0x68, 0x9f,
	};
	return ct[crc ^ data];
}

static int
ssb_sprom(struct ssb_softc *ssb, struct ssb_sprom *out,
    const uint16_t *in, uint16_t size)
{

	memset(out, 0, sizeof(*out));

	out->rev = in[size - 1] & 0x00FF;
	DPRINTF(ssb, SSB_DEBUG_SPROM, "SPROM rev %d\n",
	    out->rev);
	memset(out->mac_eth, 0xff, 6);
	memset(out->mac_80211a, 0xff, 6);
	if ((ssb->ssb_chipid & 0xff00) == 0x4400) {
		out->rev = 1;
		ssb_sprom_r123(out, in);
	} else if (ssb->ssb_chipid == 0x4321) {
		out->rev = 4;
		ssb_sprom_r45(out, in);
	} else {
		switch (out->rev) {
		case 1:
		case 2:
		case 3:
			ssb_sprom_r123(out, in);
			break;
		case 4:
		case 5:
			ssb_sprom_r45(out, in);
			break;
		case 8:
			ssb_sprom_r8(out, in);
			break;
		default:
			device_printf(ssb->ssb_dev,
			    "unknown SPROM revision %d.\n", out->rev);
			ssb_sprom_r123(out, in);
		}
	}

	if (out->bf_lo == 0xffff)
		out->bf_lo = 0;
	if (out->bf_hi == 0xffff)
		out->bf_hi = 0;

	return (0);
}

#define	SSB_LOWEST_SET_BIT(__mask) ((((__mask) - 1) & (__mask)) ^ (__mask))
#define SSB_OFFSET(offset)	(((offset) - SSB_SPROM_BASE) / sizeof(uint16_t))
#define	SSB_SHIFTOUT_SUB(__x, __mask)					\
	(((__x) & (__mask)) / SSB_LOWEST_SET_BIT(__mask))
#define	SSB_SHIFTOUT(_var, _offset, _mask)				\
	out->_var = SSB_SHIFTOUT_SUB(in[SSB_OFFSET(_offset)], (_mask))

static void
ssb_sprom_r123(struct ssb_sprom *out, const uint16_t *in)
{
	int i;
	uint16_t v;
	int8_t gain;
	uint16_t loc[3];

	if (out->rev == 3)
		loc[0] = SSB_SPROM3_MAC_80211BG;
	else {
		loc[0] = SSB_SPROM1_MAC_80211BG;
		loc[1] = SSB_SPROM1_MAC_ETH;
		loc[2] = SSB_SPROM1_MAC_80211A;
	}
	for (i = 0; i < 3; i++) {
		v = in[SSB_OFFSET(loc[0]) + i];
		*(((uint16_t *)out->mac_80211bg) + i) = htobe16(v);
	}
	if (out->rev < 3) {
		for (i = 0; i < 3; i++) {
			v = in[SSB_OFFSET(loc[1]) + i];
			*(((uint16_t *)out->mac_eth) + i) = htobe16(v);
		}
		for (i = 0; i < 3; i++) {
			v = in[SSB_OFFSET(loc[2]) + i];
			*(((uint16_t *)out->mac_80211a) + i) = htobe16(v);
		}
	}
	SSB_SHIFTOUT(mii_eth0, SSB_SPROM1_ETHPHY, SSB_SPROM1_ETHPHY_MII_ETH0);
	SSB_SHIFTOUT(mii_eth1, SSB_SPROM1_ETHPHY, SSB_SPROM1_ETHPHY_MII_ETH1);
	SSB_SHIFTOUT(mdio_eth0, SSB_SPROM1_ETHPHY, SSB_SPROM1_ETHPHY_MDIO_ETH0);
	SSB_SHIFTOUT(mdio_eth1, SSB_SPROM1_ETHPHY, SSB_SPROM1_ETHPHY_MDIO_ETH1);
	SSB_SHIFTOUT(brev, SSB_SPROM1_BOARDINFO, SSB_SPROM1_BOARDINFO_BREV);
	SSB_SHIFTOUT(ccode, SSB_SPROM1_BOARDINFO, SSB_SPROM1_BOARDINFO_CCODE);
	SSB_SHIFTOUT(ant_a, SSB_SPROM1_BOARDINFO, SSB_SPROM1_BOARDINFO_ANTA);
	SSB_SHIFTOUT(ant_bg, SSB_SPROM1_BOARDINFO, SSB_SPROM1_BOARDINFO_ANTBG);
	SSB_SHIFTOUT(pa0b0, SSB_SPROM1_PA0B0, 0xffff);
	SSB_SHIFTOUT(pa0b1, SSB_SPROM1_PA0B1, 0xffff);
	SSB_SHIFTOUT(pa0b2, SSB_SPROM1_PA0B2, 0xffff);
	SSB_SHIFTOUT(pa1b0, SSB_SPROM1_PA1B0, 0xffff);
	SSB_SHIFTOUT(pa1b1, SSB_SPROM1_PA1B1, 0xffff);
	SSB_SHIFTOUT(pa1b2, SSB_SPROM1_PA1B2, 0xffff);
	SSB_SHIFTOUT(gpio0, SSB_SPROM1_GPIOA, SSB_SPROM1_GPIOA_P0);
	SSB_SHIFTOUT(gpio1, SSB_SPROM1_GPIOA, SSB_SPROM1_GPIOA_P1);
	SSB_SHIFTOUT(gpio2, SSB_SPROM1_GPIOB, SSB_SPROM1_GPIOB_P2);
	SSB_SHIFTOUT(gpio3, SSB_SPROM1_GPIOB, SSB_SPROM1_GPIOB_P3);
	SSB_SHIFTOUT(maxpwr_a, SSB_SPROM1_MAXPWR, SSB_SPROM1_MAXPWR_A);
	SSB_SHIFTOUT(maxpwr_bg, SSB_SPROM1_MAXPWR, SSB_SPROM1_MAXPWR_BG);
	SSB_SHIFTOUT(tssi_a, SSB_SPROM1_TSSI, SSB_SPROM1_TSSI_A);
	SSB_SHIFTOUT(tssi_bg, SSB_SPROM1_TSSI, SSB_SPROM1_TSSI_BG);
	SSB_SHIFTOUT(bf_lo, SSB_SPROM1_BFLOW, 0xffff);
	if (out->rev >= 2)
		SSB_SHIFTOUT(bf_hi, SSB_SPROM2_BFHIGH, 0xffff);

	/* antenna gain */
	gain = ssb_sprom_r123_antgain(out->rev, in, SSB_SPROM1_AGAIN_BG, 0);
	out->again.ghz24.a0 = out->again.ghz24.a1 = gain;
	out->again.ghz24.a2 = out->again.ghz24.a3 = gain;
	gain = ssb_sprom_r123_antgain(out->rev, in, SSB_SPROM1_AGAIN_A, 8);
	out->again.ghz5.a0 = out->again.ghz5.a1 = gain;
	out->again.ghz5.a2 = out->again.ghz5.a3 = gain;
}

static void
ssb_sprom_r45(struct ssb_sprom *out, const uint16_t *in)
{
	int i;
	uint16_t v;
	uint16_t mac_80211bg_offset;

	if (out->rev == 4)
		mac_80211bg_offset = SSB_SPROM4_MAC_80211BG;
	else
		mac_80211bg_offset = SSB_SPROM5_MAC_80211BG;
	for (i = 0; i < 3; i++) {
		v = in[SSB_OFFSET(mac_80211bg_offset) + i];
		*(((uint16_t *)out->mac_80211bg) + i) = htobe16(v);
	}
	SSB_SHIFTOUT(mii_eth0, SSB_SPROM4_ETHPHY, SSB_SPROM4_ETHPHY_ET0A);
	SSB_SHIFTOUT(mii_eth1, SSB_SPROM4_ETHPHY, SSB_SPROM4_ETHPHY_ET1A);
	if (out->rev == 4) {
		SSB_SHIFTOUT(ccode, SSB_SPROM4_CCODE, 0xffff);
		SSB_SHIFTOUT(bf_lo, SSB_SPROM4_BFLOW, 0xffff);
		SSB_SHIFTOUT(bf_hi, SSB_SPROM4_BFHIGH, 0xffff);
	} else {
		SSB_SHIFTOUT(ccode, SSB_SPROM5_CCODE, 0xffff);
		SSB_SHIFTOUT(bf_lo, SSB_SPROM5_BFLOW, 0xffff);
		SSB_SHIFTOUT(bf_hi, SSB_SPROM5_BFHIGH, 0xffff);
	}
	SSB_SHIFTOUT(ant_a, SSB_SPROM4_ANTAVAIL, SSB_SPROM4_ANTAVAIL_A);
	SSB_SHIFTOUT(ant_bg, SSB_SPROM4_ANTAVAIL, SSB_SPROM4_ANTAVAIL_BG);
	SSB_SHIFTOUT(maxpwr_bg, SSB_SPROM4_MAXP_BG, SSB_SPROM4_MAXP_BG_MASK);
	SSB_SHIFTOUT(tssi_bg, SSB_SPROM4_MAXP_BG, SSB_SPROM4_TSSI_BG);
	SSB_SHIFTOUT(maxpwr_a, SSB_SPROM4_MAXP_A, SSB_SPROM4_MAXP_A_MASK);
	SSB_SHIFTOUT(tssi_a, SSB_SPROM4_MAXP_A, SSB_SPROM4_TSSI_A);
	if (out->rev == 4) {
		SSB_SHIFTOUT(gpio0, SSB_SPROM4_GPIOA, SSB_SPROM4_GPIOA_P0);
		SSB_SHIFTOUT(gpio1, SSB_SPROM4_GPIOA, SSB_SPROM4_GPIOA_P1);
		SSB_SHIFTOUT(gpio2, SSB_SPROM4_GPIOB, SSB_SPROM4_GPIOB_P2);
		SSB_SHIFTOUT(gpio3, SSB_SPROM4_GPIOB, SSB_SPROM4_GPIOB_P3);
	} else {
		SSB_SHIFTOUT(gpio0, SSB_SPROM5_GPIOA, SSB_SPROM5_GPIOA_P0);
		SSB_SHIFTOUT(gpio1, SSB_SPROM5_GPIOA, SSB_SPROM5_GPIOA_P1);
		SSB_SHIFTOUT(gpio2, SSB_SPROM5_GPIOB, SSB_SPROM5_GPIOB_P2);
		SSB_SHIFTOUT(gpio3, SSB_SPROM5_GPIOB, SSB_SPROM5_GPIOB_P3);
	}

	/* antenna gain */
	SSB_SHIFTOUT(again.ghz24.a0, SSB_SPROM4_AGAIN01, SSB_SPROM4_AGAIN0);
	SSB_SHIFTOUT(again.ghz24.a1, SSB_SPROM4_AGAIN01, SSB_SPROM4_AGAIN1);
	SSB_SHIFTOUT(again.ghz24.a2, SSB_SPROM4_AGAIN23, SSB_SPROM4_AGAIN2);
	SSB_SHIFTOUT(again.ghz24.a3, SSB_SPROM4_AGAIN23, SSB_SPROM4_AGAIN3);
	bcopy(&out->again.ghz24, &out->again.ghz5, sizeof(out->again.ghz5));
}

static void
ssb_sprom_r8(struct ssb_sprom *out, const uint16_t *in)
{
	int i;
	uint16_t v;

	for (i = 0; i < 3; i++) {
		v = in[SSB_OFFSET(SSB_SPROM1_MAC_80211BG) + i];
		*(((uint16_t *)out->mac_80211bg) + i) = htobe16(v);
	}
	SSB_SHIFTOUT(ccode, SSB_SPROM8_CCODE, 0xffff);
	SSB_SHIFTOUT(bf_lo, SSB_SPROM8_BFLOW, 0xffff);
	SSB_SHIFTOUT(bf_hi, SSB_SPROM8_BFHIGH, 0xffff);
	SSB_SHIFTOUT(ant_a, SSB_SPROM8_ANTAVAIL, SSB_SPROM8_ANTAVAIL_A);
	SSB_SHIFTOUT(ant_bg, SSB_SPROM8_ANTAVAIL, SSB_SPROM8_ANTAVAIL_BG);
	SSB_SHIFTOUT(maxpwr_bg, SSB_SPROM8_MAXP_BG, SSB_SPROM8_MAXP_BG_MASK);
	SSB_SHIFTOUT(tssi_bg, SSB_SPROM8_MAXP_BG, SSB_SPROM8_TSSI_BG);
	SSB_SHIFTOUT(maxpwr_a, SSB_SPROM8_MAXP_A, SSB_SPROM8_MAXP_A_MASK);
	SSB_SHIFTOUT(tssi_a, SSB_SPROM8_MAXP_A, SSB_SPROM8_TSSI_A);
	SSB_SHIFTOUT(gpio0, SSB_SPROM8_GPIOA, SSB_SPROM8_GPIOA_P0);
	SSB_SHIFTOUT(gpio1, SSB_SPROM8_GPIOA, SSB_SPROM8_GPIOA_P1);
	SSB_SHIFTOUT(gpio2, SSB_SPROM8_GPIOB, SSB_SPROM8_GPIOB_P2);
	SSB_SHIFTOUT(gpio3, SSB_SPROM8_GPIOB, SSB_SPROM8_GPIOB_P3);

	/* antenna gain */
	SSB_SHIFTOUT(again.ghz24.a0, SSB_SPROM8_AGAIN01, SSB_SPROM8_AGAIN0);
	SSB_SHIFTOUT(again.ghz24.a1, SSB_SPROM8_AGAIN01, SSB_SPROM8_AGAIN1);
	SSB_SHIFTOUT(again.ghz24.a2, SSB_SPROM8_AGAIN23, SSB_SPROM8_AGAIN2);
	SSB_SHIFTOUT(again.ghz24.a3, SSB_SPROM8_AGAIN23, SSB_SPROM8_AGAIN3);
	bcopy(&out->again.ghz24, &out->again.ghz5, sizeof(out->again.ghz5));
}

static int8_t
ssb_sprom_r123_antgain(uint8_t sprom_revision, const uint16_t *in,
    uint16_t mask, uint16_t shift)
{
	uint16_t v;
	uint8_t gain;

	v = in[SSB_OFFSET(SSB_SPROM1_AGAIN)];
	gain = (v & mask) >> shift;
	gain = (gain == 0xff) ? 2 : (sprom_revision == 1) ? gain << 2 :
	    ((gain & 0xc0) >> 6) | ((gain & 0x3f) << 2);

	return (int8_t)gain;
}

#undef SSB_LOWEST_SET_BIT
#undef SSB_OFFSET
#undef SSB_SHIFTOUT_SUB
#undef SSB_SHIFTOUT

int
ssb_powerdown(struct ssb_softc *ssb)
{
	struct ssb_cc *scc;

	if (ssb->ssb_type == SSB_TYPE_SSB)
		return (0);

	scc = &ssb->ssb_cc;
	if (!scc->scc_dev || scc->scc_dev->sd_id.rev < 5)
		return (0);
	ssb_cc_clock(scc, SSB_CLOCK_SLOW);
	ssb_pci_gpio(ssb, SSB_GPIO_CRYSTAL | SSB_GPIO_PLL, 0);
	return (0);
}

static void
ssb_pcicore_init(struct ssb_pci *spc)
{
	struct ssb_dev_softc *sd = spc->spc_dev;
	struct ssb_softc *ssb;

	if (sd == NULL)
		return;

	ssb = sd->sd_bus;
	if (!ssb_dev_isup(sd))
		ssb_dev_up(sd, 0);

	KASSERT(spc->spc_hostmode == 0,
	    ("%s:%d: hostmode", __func__, __LINE__));
	/* disable PCI interrupt */
	ssb_write_4(spc->spc_dev, SSB_INTR_MASK, 0);
}

int
ssb_dev_isup(struct ssb_dev_softc *sd)
{
	uint32_t reject, val;

	reject = ssb_tmslow_reject_bitmask(sd);
	val = ssb_read_4(sd, SSB_TGSLOW);
	val &= SSB_TGSLOW_CLOCK | SSB_TGSLOW_RESET | reject;

	return (val == SSB_TGSLOW_CLOCK);
}

void
ssb_dev_up(struct ssb_dev_softc *sd, uint32_t flags)
{
	uint32_t val;

	ssb_dev_down(sd, flags);
	ssb_write_4(sd, SSB_TGSLOW, SSB_TGSLOW_RESET | SSB_TGSLOW_CLOCK |
	    SSB_TGSLOW_FGC | flags);
	ssb_read_4(sd, SSB_TGSLOW);
	DELAY(1);

	if (ssb_read_4(sd, SSB_TGSHIGH) & SSB_TGSHIGH_SERR)
		ssb_write_4(sd, SSB_TGSHIGH, 0);

	val = ssb_read_4(sd, SSB_IAS);
	if (val & (SSB_IAS_INBAND_ERR | SSB_IAS_TIMEOUT)) {
		val &= ~(SSB_IAS_INBAND_ERR | SSB_IAS_TIMEOUT);
		ssb_write_4(sd, SSB_IAS, val);
	}

	ssb_write_4(sd, SSB_TGSLOW, SSB_TGSLOW_CLOCK | SSB_TGSLOW_FGC | flags);
	ssb_read_4(sd, SSB_TGSLOW);
	DELAY(1);

	ssb_write_4(sd, SSB_TGSLOW, SSB_TGSLOW_CLOCK | flags);
	ssb_read_4(sd, SSB_TGSLOW);
	DELAY(1);
}

static uint32_t
ssb_tmslow_reject_bitmask(struct ssb_dev_softc *sd)
{
	uint32_t rev = ssb_read_4(sd, SSB_IDLOW) & SSB_IDLOW_SSBREV;

	switch (rev) {
	case SSB_IDLOW_SSBREV_22:
		return SSB_TGSLOW_REJECT_22;
	case SSB_IDLOW_SSBREV_23:
		return SSB_TGSLOW_REJECT_23;
	case SSB_IDLOW_SSBREV_24:
	case SSB_IDLOW_SSBREV_25:
	case SSB_IDLOW_SSBREV_26:
	case SSB_IDLOW_SSBREV_27:
		return SSB_TGSLOW_REJECT_23;
	default:
		KASSERT(0 == 1,
		    ("%s:%d: unknown backplane rev %#x\n",
			__func__, __LINE__, rev));
	}
	return (SSB_TGSLOW_REJECT_22 | SSB_TGSLOW_REJECT_23);
}

void
ssb_dev_down(struct ssb_dev_softc *sd, uint32_t flags)
{
	struct ssb_softc *ssb = sd->sd_bus;
	uint32_t reject, val;
	int i;

	if (ssb_read_4(sd, SSB_TGSLOW) & SSB_TGSLOW_RESET)
		return;

	reject = ssb_tmslow_reject_bitmask(sd);
	ssb_write_4(sd, SSB_TGSLOW, reject | SSB_TGSLOW_CLOCK);

	for (i = 0; i < 1000; i++) {
		val = ssb_read_4(sd, SSB_TGSLOW);
		if (val & reject)
			break;
		DELAY(10);
	}
	if ((val & reject) == 0) {
		device_printf(ssb->ssb_dev, "timeout (bit %#x reg %#x)\n",
		    reject, SSB_TGSLOW);
	}
	for (i = 0; i < 1000; i++) {
		val = ssb_read_4(sd, SSB_TGSHIGH);
		if (!(val & SSB_TGSHIGH_BUSY))
			break;
		DELAY(10);
	}
	if ((val & SSB_TGSHIGH_BUSY) != 0) {
		device_printf(ssb->ssb_dev, "timeout (bit %#x reg %#x)\n",
		    SSB_TGSHIGH_BUSY, SSB_TGSHIGH);
	}

	ssb_write_4(sd, SSB_TGSLOW, SSB_TGSLOW_FGC | SSB_TGSLOW_CLOCK |
	    reject | SSB_TGSLOW_RESET | flags);
	ssb_read_4(sd, SSB_TGSLOW);
	DELAY(1);
	ssb_write_4(sd, SSB_TGSLOW, reject | SSB_TGSLOW_RESET | flags);
	ssb_read_4(sd, SSB_TGSLOW);
	DELAY(1);
}

void
ssb_pcicore_intr(struct ssb_pci *spc, struct ssb_dev_softc *sd)
{
	struct ssb_dev_softc *psd = spc->spc_dev;
	struct ssb_softc *ssb;
	uint32_t tmp;

	if (sd->sd_bus->ssb_type != SSB_TYPE_PCI || !psd)
		return;

	ssb = psd->sd_bus;
	/* enable interrupts */
	if (ssb->ssb_dev != NULL &&
	    (psd->sd_id.rev >= 6 || psd->sd_id.cid == SSB_DEV_PCIE)) {
		tmp = pci_read_config(ssb->ssb_dev, SSB_IRQMASK, 4);
		tmp |= (1 << sd->sd_coreidx) << 8;
		pci_write_config(ssb->ssb_dev, SSB_IRQMASK, tmp, 4);
	} else {
		tmp = ssb_read_4(sd, SSB_TPS);
		tmp &= SSB_TPS_BPFLAG;
		ssb_write_4(psd, SSB_INTR_MASK,
		    ssb_read_4(psd, SSB_INTR_MASK) | (1 << tmp));
	}

	/* setup PCIcore */
	if (spc->spc_inited == 1)
		return;
	if (psd->sd_id.cid == SSB_DEV_PCI) {
		ssb_pcicore_write32(spc, SSB_PCICORE_SBTOPCI2,
		    ssb_pcicore_read32(spc, SSB_PCICORE_SBTOPCI2) |
		    SSB_PCICORE_SBTOPCI_PREF | SSB_PCICORE_SBTOPCI_BURST);

		if (psd->sd_id.rev < 5) {
			tmp = ssb_read_4(psd, SSB_IMCFGLO);
			tmp &= ~SSB_IMCFGLO_SERTO;
			tmp = (tmp | 2) & ~SSB_IMCFGLO_REQTO;
			tmp |= 3 << 4 /* SSB_IMCFGLO_REQTO_SHIFT */;
			ssb_write_4(psd, SSB_IMCFGLO, tmp);

			/* broadcast value */
			sd = (ssb->ssb_cc.scc_dev != NULL) ?
			    ssb->ssb_cc.scc_dev : ssb->ssb_pci.spc_dev;
			if (sd != NULL) {
				ssb_write_4(sd, SSB_PCICORE_BCAST_ADDR, 0xfd8);
				ssb_read_4(sd, SSB_PCICORE_BCAST_ADDR);
				ssb_write_4(sd, SSB_PCICORE_BCAST_DATA, 0);
				ssb_read_4(sd, SSB_PCICORE_BCAST_DATA);
			}
		} else if (psd->sd_id.rev >= 11) {
			tmp = ssb_pcicore_read32(spc, SSB_PCICORE_SBTOPCI2);
			tmp |= SSB_PCICORE_SBTOPCI_MRM;
			ssb_pcicore_write32(spc, SSB_PCICORE_SBTOPCI2, tmp);
		}
	} else {
		KASSERT(psd->sd_id.cid == SSB_DEV_PCIE, ("only PCIE"));
		if ((psd->sd_id.rev == 0) || (psd->sd_id.rev == 1))
			ssb_pcie_write(spc, 0x4, ssb_pcie_read(spc, 0x4) | 0x8);
		if (psd->sd_id.rev == 0) {
			ssb_pcie_mdio_write(spc, 0x1f, 2, 0x8128); /* Timer */
			ssb_pcie_mdio_write(spc, 0x1f, 6, 0x0100); /* CDR */
			ssb_pcie_mdio_write(spc, 0x1f, 7, 0x1466); /* CDR BW */
		} else if (psd->sd_id.rev == 1)
			ssb_pcie_write(spc, 0x100,
			    ssb_pcie_read(spc, 0x100) | 0x40);
	}
	spc->spc_inited = 1;
}

static uint32_t
ssb_pcicore_read32(struct ssb_pci *spc, uint16_t offset)
{

	return ssb_read_4(spc->spc_dev, offset);
}

static void
ssb_pcicore_write32(struct ssb_pci *spc, uint16_t offset, uint32_t value)
{

	ssb_write_4(spc->spc_dev, offset, value);
}

static uint32_t
ssb_pcie_read(struct ssb_pci *spc, uint32_t address)
{

	ssb_pcicore_write32(spc, 0x130, address);
	return ssb_pcicore_read32(spc, 0x134);
}

static void
ssb_pcie_write(struct ssb_pci *spc, uint32_t address, uint32_t data)
{

	ssb_pcicore_write32(spc, 0x130, address);
	ssb_pcicore_write32(spc, 0x134, data);
}

static void
ssb_pcie_mdio_write(struct ssb_pci *spc, uint8_t device, uint8_t address,
    uint16_t data)
{
	int i;

	ssb_pcicore_write32(spc, SSB_PCICORE_MDIO_CTL, 0x80 | 0x2);
	ssb_pcicore_write32(spc, SSB_PCICORE_MDIO_DATA, (1 << 30) | (1 << 28) |
	    ((uint32_t)device << 22) | ((uint32_t)address << 18) |
	    (1 << 17) | data);
	DELAY(10);
	for (i = 0; i < 10; i++) {
		if (ssb_pcicore_read32(spc, SSB_PCICORE_MDIO_CTL) & 0x100)
			break;
		DELAY(1000);
	}
	ssb_pcicore_write32(spc, SSB_PCICORE_MDIO_CTL, 0);
}

uint32_t
ssb_dma_translation(struct ssb_dev_softc *sd)
{

	KASSERT(sd->sd_bus->ssb_type == SSB_TYPE_PCI,
	    ("unsupported bustype %d\n", sd->sd_bus->ssb_type));
	return (SSB_PCI_DMA);
}

void
ssb_barrier(struct ssb_dev_softc *sd, int flags)
{
	struct ssb_softc *ssb = sd->sd_bus;

	SSB_BARRIER(ssb, flags);
}

/*
 * Attach it as child.
 */
device_t
ssb_add_child(device_t dev, struct ssb_softc *ssb, int order,
    const char *name, int unit)
{
	struct ssb_dev_softc *sd;
        device_t child;
	int idx = 0, i;

	child = device_add_child(dev, name, unit);
	if (child == NULL)
		return (NULL);

	ssb_powerup(ssb, 0);
	ssb_pcicore_init(&ssb->ssb_pci);
	ssb_powerdown(ssb);

	for (i = 0; i < ssb->ssb_ndevs; i++) {
		sd = &(ssb->ssb_devs[i]);

		switch (sd->sd_id.cid) {
		case SSB_DEV_CHIPCOMMON:
		case SSB_DEV_PCI:
		case SSB_DEV_PCIE:
		case SSB_DEV_PCMCIA:
		case SSB_DEV_MIPS:
		case SSB_DEV_MIPS_3302:
		case SSB_DEV_V90:
		case SSB_DEV_EXTIF:
			DPRINTF(ssb, SSB_DEBUG_CORE,
			    "skip to register coreid %#x (%s)\n",
			    sd->sd_id.cid, ssb_core_name(sd->sd_id.cid));
			continue;
		}

		DPRINTF(ssb, SSB_DEBUG_CORE,
		    "ssb: attaching coreid %#x (%s) idx %d\n",
		    sd->sd_id.cid, ssb_core_name(sd->sd_id.cid), idx);

		KASSERT(sd->sd_id.cid == SSB_DEV_80211,
		    ("%s:%d: SSB_DEV_80211 is only supportted currently.",
			__func__, __LINE__));

		device_set_ivars(child, sd);
		device_probe_and_attach(child);
		idx++;
	}
	return (child);
}
