/*-
 * Copyright (c) 2008 Weongyo Jeong <weongyo@freebsd.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer,
 *    without modification.
 * 2. Redistributions in binary form must reproduce at minimum a disclaimer
 *    similar to the "NO WARRANTY" disclaimer below ("Disclaimer") and any
 *    redistribution must be conditioned upon including a substantially
 *    similar Disclaimer requirement for further binary redistribution.
 *
 * NO WARRANTY
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF NONINFRINGEMENT, MERCHANTIBILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGES.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

/*
 * PCI/Cardbus front-end for the Sonics Silicon Backplane(SSB) driver.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/module.h>
#include <sys/kernel.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/errno.h>
#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/bus.h>
#include <sys/rman.h>
#include <sys/socket.h>

#include <net/if.h>
#include <net/if_media.h>
#include <net/if_arp.h>

#include <dev/pci/pcivar.h>
#include <dev/pci/pcireg.h>
#include <dev/ssb/ssb.h>

/*
 * PCI glue.
 */

struct ssb_pci_softc {
	struct ssb_softc		psc_ssb;
};

#define	BS_BAR				0x10
#define	PCI_VENDOR_BROADCOM		0x14e4
#define	N(a)				(sizeof(a) / sizeof(a[0]))

static const struct ssb_dev {
	uint16_t	vid;
	uint16_t	did;
	const char	*desc;
} ssb_devices[] = {
	{ PCI_VENDOR_BROADCOM, 0x4301, "Broadcom BCM4301 802.11b Wireless" },
	{ PCI_VENDOR_BROADCOM, 0x4306, "Unknown" },
	{ PCI_VENDOR_BROADCOM, 0x4307, "Broadcom BCM4307 802.11b Wireless" },
	{ PCI_VENDOR_BROADCOM, 0x4311, "Broadcom BCM4311 802.11b/g Wireless" },
	{ PCI_VENDOR_BROADCOM, 0x4312,
	  "Broadcom BCM4312 802.11a/b/g Wireless" },
	{ PCI_VENDOR_BROADCOM, 0x4315, "Broadcom BCM4315 802.11b/g Wireless" },
	{ PCI_VENDOR_BROADCOM, 0x4318, "Broadcom BCM4318 802.11b/g Wireless" },
	{ PCI_VENDOR_BROADCOM, 0x4319,
	  "Broadcom BCM4318 802.11a/b/g Wireless" },
	{ PCI_VENDOR_BROADCOM, 0x4320, "Broadcom BCM4306 802.11b/g Wireless" },
	{ PCI_VENDOR_BROADCOM, 0x4321, "Broadcom BCM4306 802.11a Wireless" },
	{ PCI_VENDOR_BROADCOM, 0x4324,
	  "Broadcom BCM4309 802.11a/b/g Wireless" },
	{ PCI_VENDOR_BROADCOM, 0x4325, "Broadcom BCM4306 802.11b/g Wireless" },
	{ PCI_VENDOR_BROADCOM, 0x4328, "Unknown" },
	{ PCI_VENDOR_BROADCOM, 0x4329, "Unknown" },
	{ PCI_VENDOR_BROADCOM, 0x432b, "Unknown" }
};

device_t	ssb_add_child(device_t, struct ssb_softc *, int, const char *,
		    int);

static int
ssb_pci_probe(device_t dev)
{
	int i;
	uint16_t did, vid;

	did = pci_get_device(dev);
	vid = pci_get_vendor(dev);

	for (i = 0; i < N(ssb_devices); i++) {
		if (ssb_devices[i].did == did && ssb_devices[i].vid == vid) {
			device_set_desc(dev, ssb_devices[i].desc);
			return (BUS_PROBE_DEFAULT);
		}
	}
	return (ENXIO);
}

static int
ssb_pci_attach(device_t dev)
{
	struct ssb_pci_softc *psc = device_get_softc(dev);
	struct ssb_softc *ssb = &psc->psc_ssb;

	ssb->ssb_dev = dev;
	ssb->ssb_type = SSB_TYPE_PCI;

	/*
	 * Enable bus mastering.
	 */
	pci_enable_busmaster(dev);

	/* 
	 * Setup memory-mapping of PCI registers.
	 */
	ssb->ssb_mem_rid = SSB_PCIR_BAR;
	ssb->ssb_mem_res = bus_alloc_resource_any(dev, SYS_RES_MEMORY,
		&ssb->ssb_mem_rid, RF_ACTIVE);
	if (ssb->ssb_mem_res == NULL) {
		device_printf(dev, "cannot map register space\n");
		return (ENXIO);
	}
	ssb->ssb_mem_bt = rman_get_bustag(ssb->ssb_mem_res);
	ssb->ssb_mem_bh = rman_get_bushandle(ssb->ssb_mem_res);

	/* Get more PCI information */
	ssb->ssb_pci_did = pci_get_device(dev);
	ssb->ssb_pci_vid = pci_get_vendor(dev);
	ssb->ssb_pci_subvid = pci_get_subvendor(dev);
	ssb->ssb_pci_subdid = pci_get_subdevice(dev);

	return (ssb_attach(ssb));
}

static int
ssb_pci_detach(device_t dev)
{
	struct ssb_pci_softc *psc = device_get_softc(dev);
	struct ssb_softc *ssb = &psc->psc_ssb;

	/* check if device was removed */
	ssb->ssb_invalid = !bus_child_present(dev);

	pci_disable_busmaster(dev);
	bus_generic_detach(dev);
	ssb_detach(ssb);
	bus_release_resource(dev, SYS_RES_MEMORY, BS_BAR, ssb->ssb_mem_res);

	return (0);
}

static int
ssb_pci_shutdown(device_t dev)
{

	return (ENXIO);
}

static int
ssb_pci_suspend(device_t dev)
{

	return (ENXIO);
}

static int
ssb_pci_resume(device_t dev)
{

	return (ENXIO);
}

static device_t
ssb_pci_add_child(device_t dev, int order, const char *name, int unit)
{
	struct ssb_pci_softc *psc = device_get_softc(dev);
	struct ssb_softc *ssb = &psc->psc_ssb;

	return (ssb_add_child(dev, ssb, order, name, unit));
}

static struct resource *
ssb_pci_alloc_resource(device_t dev, device_t child, int type, int *rid,
    u_long start, u_long end, u_long count, u_int flags)
{

	return (BUS_ALLOC_RESOURCE(device_get_parent(dev), dev,
	    type, rid, start, end, count, flags));
}

static int
ssb_pci_release_resource(device_t dev, device_t child, int type,
    int rid, struct resource *r)
{

	return (BUS_RELEASE_RESOURCE(device_get_parent(dev), dev, type,
	    rid, r));
}

static int
ssb_pci_setup_intr(device_t dev, device_t child, struct resource *irq,
    int flags, driver_filter_t *filter, driver_intr_t *intr, void *arg,
    void **cookiep)
{

	return (BUS_SETUP_INTR(device_get_parent(dev), dev, irq, flags,
	    filter, intr, arg, cookiep));
}

static int
ssb_pci_teardown_intr(device_t dev, device_t child, struct resource *irq,
    void *cookie)
{

	return (BUS_TEARDOWN_INTR(device_get_parent(dev), dev, irq, cookie));
}

static device_method_t ssb_pci_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe,		ssb_pci_probe),
	DEVMETHOD(device_attach,	ssb_pci_attach),
	DEVMETHOD(device_detach,	ssb_pci_detach),
	DEVMETHOD(device_shutdown,	ssb_pci_shutdown),
	DEVMETHOD(device_suspend,	ssb_pci_suspend),
	DEVMETHOD(device_resume,	ssb_pci_resume),

	/* Bus interface */
	DEVMETHOD(bus_add_child,	ssb_pci_add_child),
	DEVMETHOD(bus_alloc_resource,   ssb_pci_alloc_resource),
	DEVMETHOD(bus_release_resource, ssb_pci_release_resource),
	DEVMETHOD(bus_setup_intr,       ssb_pci_setup_intr),
        DEVMETHOD(bus_teardown_intr,    ssb_pci_teardown_intr),

	{ 0,0 }
};
static driver_t ssb_pci_driver = {
	"ssb",
	ssb_pci_methods,
	sizeof(struct ssb_pci_softc)
};
static devclass_t ssb_pci_devclass;
DRIVER_MODULE(ssb, pci, ssb_pci_driver, ssb_pci_devclass, 0, 0);
MODULE_VERSION(ssb, 1);
